<?php
// Own namespace
namespace Org\Mxchange\CoreFramework\Registry;

// Import framework stuff
use Org\Mxchange\CoreFramework\Factory\ObjectFactory;
use Org\Mxchange\CoreFramework\Generic\NullPointerException;
use Org\Mxchange\CoreFramework\Object\BaseFrameworkSystem;

// Import SPL stuff
use \IteratorAggregate;

/**
 * A general Registry
 *
 * @author		Roland Haeder <webmaster@shipsimu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2017 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.shipsimu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
class BaseRegistry extends BaseFrameworkSystem implements Register, Registerable, IteratorAggregate {
	/**
	 * Glue for generating a registry key
	 */
	const REGISTRY_KEY_GLUE = '_';

	/**
	 * Instance of this class
	 */
	private static $registryInstance = NULL;

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
	 * @return	void
	 */
	protected function __construct ($className) {
		// Call parent constructor
		parent::__construct($className);

		// Init generic arrays
		$this->initGenericArrayGroup('registry', 'generic');
		$this->initGenericArrayGroup('registry', 'instance');
	}

	/**
	 * Returns an iterator for this whole registry.
	 *
	 * @return	$iteratorInstance	An instance of a Iterator class
	 */
	public function getIterator () {
		// Is it set?
		if (is_null($this->getIteratorInstance())) {
			// Then instance it
			$iteratorInstance = ObjectFactory::createObjectByConfiguredName('registry_iterator_class', array($this));

			// ... and set it here
			$this->setIteratorInstance($iteratorInstance);
		} else {
			// Use set iterator
			$iteratorInstance = $this->getIteratorInstance();
		}

		// Return it
		return $iteratorInstance;
	}

	/**
	 * Checks whether an instance key was found
	 *
	 * @param	$instanceKey	The key holding an instance in registry
	 * @return	$exists			Whether the key exists in registry
	 */
	public function instanceExists ($instanceKey) {
		// Does this key exists?
		$exists = $this->isGenericArrayKeySet('registry', 'instance', $instanceKey);

		// Return the result
		return $exists;
	}

	/**
	 * Adds/overwrites a new instance to the registry at the given key
	 *
	 * @param	$instanceKey		The key to identify the instance
	 * @param	$objectInstance		An instance we shall store
	 * @return	void
	 */
	public function addInstance ($instanceKey, Registerable $objectInstance) {
		$this->setGenericArrayKey('registry', 'instance', $instanceKey, $objectInstance);
	}

	/**
	 * Getter for whole generic registry
	 *
	 * @return	$instanceRegistry	The whole generic registry array
	 */
	public final function getGenericRegistry () {
		return $this->getGenericSubArray('registry', 'generic');
	}

	/**
	 * Getter for whole instance registry
	 *
	 * @return	$instanceRegistry	The whole instance registry array
	 */
	public final function getInstanceRegistry () {
		return $this->getGenericSubArray('registry', 'instance');
	}

	/**
	 * Adds a new entry to the given list name. If you want to add objects
	 * please use addInstance() and getInstance() instead.
	 *
	 * @param	$key	The key to identify the whole list
	 * @param	$value	The value to be stored
	 * @return	void
	 */
	public final function addEntry ($key, $value) {
		// Key must not be an array
		assert(!is_array($key));

		// Push it
		$this->pushValueToGenericArrayKey('registry', 'generic', $key, $value);
	}

	/**
	 * Getter for entries or "sub entries"
	 *
	 * @return	$entries	An array with entries from this registry
	 */
	public final function getEntries ($key = NULL) {
		// Key must not be an array
		assert(!is_array($key));

		// Default is whole array
		$entries = $this->getGenericArray('registry');

		// Is $key set?
		if (!is_null($key)) {
			// Then use this entry
			$entries = $this->getGenericArrayKey('registry', 'generic', $key);
		} // END - if

		// Return the array
		return $entries;
	}

	/**
	 * "Getter" for an array of all entries for given key
	 *
	 * @param	$arrayKey	The array (key) to look in
	 * @param	$lookFor	The key to look for
	 * @return	$entry		An array with all keys
	 */
	public function getArrayFromKey ($arrayKey, $lookFor) {
		// Key must not be an array
		assert(!is_array($arrayKey));

		// Init array
		$entry = array();

		// Debug message
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('REGISTRY[' . __METHOD__ . ':' . __LINE__ . ']: Checking arrayKey=' . $arrayKey . ',lookFor=' . $lookFor);

		// "Walk" over all entries
		foreach ($this->getEntries($arrayKey) as $key => $value) {
			// Debug message
			//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('REGISTRY[' . __METHOD__ . ':' . __LINE__ . ']: Checking key=' . $key . ',value=' . $value . ',lookFor=' . $lookFor);

			// If $value matches the $lookFor, we need to look for more entries for this!
			if ($lookFor == $value) {
				// Look for more entries
				foreach ($this->getEntries() as $key2 => $value2) {
					// Now "walk" through all entries, if an array is returned
					if (is_array($value2)) {
						// Debug message
						//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('REGISTRY[' . __METHOD__ . ':' . __LINE__ . ']: Checking key2=' . $key2 . ',value2()=' . count($value2) . ',lookFor=' . $lookFor);

						// "Walk" through all of them
						foreach ($value2 as $key3 => $value3) {
							// $value3 needs to be an array
							assert(is_array($value3));

							// Debug message
							//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('REGISTRY[' . __METHOD__ . ':' . __LINE__ . ']: Checking key=' . $key . ',key3=' . $key3 . ',isset()=' . isset($value3[$key]) . ' ...');

							// Both keys must match!
							if (($key == $key3) || (isset($value3[$key]))) {
								// Debug message
								//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('REGISTRY[' . __METHOD__ . ':' . __LINE__ . ']: Adding ' . $value3[$key] . ' ...');

								// Then add it
								$entry[$key3] = $value3[$key];
							} // END - if
						} // END - foreach
					} // END - if
				} // END - foreach

				// Skip further lookups
				break;
			} // END - if
		} // END - foreach

		// Debug message
		//* DEBUG: */ self::createDebugInstance(__CLASS__, __LINE__)->debugOutput('REGISTRY[' . __METHOD__ . ':' . __LINE__ . ']: Returning entry(' . count($entry) . ')=' . print_r($entry, true));

		// Return it
		return $entry;
	}

	/**
	 * Gets a registered instance or null if not found
	 *
	 * @param	$instanceKey		The key to identify the instance
	 * @return	$objectInstance		An instance we shall store
	 * @throws	NullPointerException	If the requested key is not found
	 */
	public function getInstance ($instanceKey) {
		// Is the instance there?
		if (!$this->instanceExists($instanceKey)) {
			// This might happen if a non-registered key was requested
			throw new NullPointerException($this, self::EXCEPTION_IS_NULL_POINTER);
		} // END - if

		// Get the instance
		$objectInstance = $this->getGenericArrayKey('registry', 'instance', $instanceKey);

		// Return the result
		return $objectInstance;
	}

	/**
	 * "Getter" for a registry key for given prefix and array. This method
	 * calls implode() to get a suitable key. This method does not care about
	 * the indexes.
	 *
	 * @param	$prefix			Prefix for the key
	 * @param	$data			An array with data
	 * @return	$registryKey	A registry key
	 */
	public static function getRegistryKeyFromArray ($prefix, array $data) {
		// "Generate" the key
		$registryKey = $prefix . self::REGISTRY_KEY_GLUE . implode(self::REGISTRY_KEY_GLUE, $data);

		// Return it
		return $registryKey;
	}

}
